/*

 normal.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "normal.h"
#include "driver.h"
#include "path.h"
#include "crc8.h"
#include "partition.h"
#include "log.h"
#include "arguments.h"
#include "buffer.h"
#include "ideservd.h"
#include "directory.h"
#include "utf8.h"
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

static int send_trailer(const Driver *driver, int usecrc) {
    if (driver->mode != M_ETHERNET) {
        if (usecrc) {
            driver->sendb((unsigned char)crc_get());
        }
        driver->sendb(0x5a);
    }
    return driver->flush() != DRIVER_SUCCESS;
}

static int frame_error(const char *mode) {
    seterror(ER_FRAME_ERROR, 1);
    log_printf("%s: Frame error", mode);
    return 1;
}

static int crc_error(const char *mode) {
    seterror(ER_CRC_ERROR, 1);
    log_printf("%s: CRC error", mode);
    return 1;
}

static Errorcode directory_error(const char *path) {
    Errorcode status = errnotochannel15(errno);
    log_printf("Open: Couldn't open the directory \"%s\": %s(%d)", path, strerror(errno), errno);
    return status;
}

int statuserror(const Driver *driver, unsigned int flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    const int debug = flags & FLAG_DEBUG;
    Petscii cmd[256];
    struct timeval start;
    unsigned char fp;
    int r;
    Buffer *buffer;

    r = driver->getb(1);
    if (usecrc) crc_add_byte(r);
    fp = 0;
    while (fp < 255) {
        r = driver->getb(0);
        if (r == EOF) break;
        cmd[fp] = (unsigned char)r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (driver->mode != M_ETHERNET) {
        unsigned char buf[2];
        if (usecrc) driver->getbytes(buf, 2);
        else buf[1] = driver->getb(0);
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (buf[1] != 0x5a || r != 0) return frame_error("Status");
        if (usecrc) {
            crc_add_block(cmd, fp + 1);
            crc_add_byte(buf[0]);
            if (crc_get()) return crc_error("Status");
        }
    } else {
        if (driver->done() != DRIVER_SUCCESS) return 1;
    }

    if (verbose) {
        gettimeofday(&start, NULL);
    }

    if (debug) {
        log_print("Status:");
        log_hex(cmd);
    }
    commandchannel(cmd);
    if (verbose) {
        log_time("Status: Took", &start);
    }

    buffer = buffers_get(15);
    fp = buffer->size <= 255 ? (unsigned char)buffer->size : 255;
    driver->sendb(fp);
    if (driver->mode == M_ETHERNET) driver->sendb(fp);
    driver->sendbytes(buffer->data, fp);
    if (usecrc) {
        crc_first(fp);
        crc_add_block(buffer->data, fp);
    }
    return send_trailer(driver, usecrc);
}

int openfile(const Driver *driver, const Arguments *arguments, unsigned int flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    const int debug = flags & FLAG_DEBUG;
    Petscii cmd[256];
    struct timeval start;
    Errorcode status = ER_OK;
    Petscii mode = 'R';
    unsigned int channel, fp;
    unsigned int length = 0;
    Path outpath;
    Buffer *buffer;
    int r;

    r = driver->getb(1);
    if (usecrc) crc_add_byte(r);
    channel = r & 0x0f;
    buffer = buffers_get(channel);

    fp = 0;
    while (fp < 255) {
        r = driver->getb(0);
        if (r == EOF) break;
        cmd[fp] = (unsigned char)r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (driver->mode != M_ETHERNET) {
        unsigned char buf[2];
        if (usecrc) driver->getbytes(buf, 2);
        else buf[1] = driver->getb(0);
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (buf[1] != 0x5a || r != 0) return frame_error("Open");
        if (usecrc) {
            crc_add_block(cmd, fp + 1);
            crc_add_byte(buf[0]);
            if (crc_get()) return crc_error("Open");
        }
    } else {
        if (driver->done() != DRIVER_SUCCESS) return 1;
    }

    if (verbose) {
        gettimeofday(&start, NULL);
    }

    if (debug) {
        log_print("Open:");
        log_hex(cmd);
    }
    buffer_close(buffer);
    buffer->filepos = 0;

    path_reset(&outpath);
    if (channel == 15) {
        buffer->mode = CM_ERR;
    } else if (cmd[0] == '$') {
        int partitionlist = 0;
        const Petscii *outname;
        const char *path = NULL;
        partition_t partition = 0;
        Partition *outpart;

        buffer->mode = CM_DIR;
        buffer->size = buffer->pointer = 0;
        if (cmd[1] >= '0' && cmd[1] <= '9') {
            int p = 1;
            while (cmd[p] >= '0' && cmd[p] <= '9') partition = partition * 10 + cmd[p++] - '0';
            if (cmd[p] == 0) cmd[1] = 0; else partition = 0;
        } 

        if (cmd[1] == 0) {
            partition_update();
            outpart = partition_get(partition);
            if (outpart == NULL) goto partnf;
            outname = cmd + 1;
            path = path_new(&outpath, outpart->path, 0);
            if (path == NULL) {
                log_print("Open: Out of memory");
                goto partnf;
            }
        } else if (cmd[1] == '=' && cmd[2] == 'P') {
            partition_update();
            path = "=p";
            partitionlist = 1;
            outname = NULL;
        } else {
            outname = resolv_part(cmd + 1, &outpath, &outpart, 0);
            if (outname == NULL) {
            partnf:
                if (verbose) log_printf("Open #%u: Partition not found", channel);
                status = ER_DRIVE_NOT_READY; goto vege;
            }
            outname = resolv_path(outname, &outpath, arguments->nameconversion);
            path = outpath.data;
            if (outname == NULL) {
                status = ER_PATH_NOT_FOUND; goto vege3;
            }
        }

        if (channel == 1) {
            status = ER_NO_CHANNEL; goto vege3;
        } else if (channel > 1) {
            Directory *directory = directory_open(&outpath, arguments->nameconversion, 3);
            if (directory == NULL) {
                status = directory_error(outpath.data);
                goto vege;
            }
            if (buffer_rawdir(buffer, directory)) {
                buffer->size &= ~511u;
                status = ER_READ_ERROR;
                log_print("Open: Out of memory");
                directory_close(directory);
            } else if (directory_close(directory)) {
                status = errnotochannel15(errno);
            } else {
                seterror(ER_OK, 0);
            }
        } else if (partitionlist) {
            if (buffer_partition(buffer, arguments->default_partition)) {
                buffer->size &= ~511u;
                status = ER_READ_ERROR;
                log_print("Open: Out of memory");
            } else {
                seterror(ER_OK, 0);
            }
        } else {
            Directory *directory = directory_open(&outpath, arguments->nameconversion, 2);

            if (directory == NULL) {
                status = directory_error(outpath.data);
                goto vege;
            }
            if (buffer_cookeddir(buffer, directory, outpart->nr, outname)) {
                buffer->size &= ~511u;
                status = ER_READ_ERROR;
                log_print("Open: Out of memory");
                directory_close(directory);
            } else if (directory_close(directory)) {
                status = errnotochannel15(errno);
            } else {
                seterror(ER_OK, 0);
            }
        }
        if ((buffer->size & 511) != 0) {
            unsigned int padding = (-buffer->size) & 511;
            if (buffer_reserve(buffer, buffer->size + padding)) {
                buffer->size &= ~511u;
                status = ER_READ_ERROR;
                log_print("Open: Out of memory");
            } else {
                memset(buffer->data + buffer->size, 0, padding);
            }
        }
        length = buffer->size;
    vege3:
        if (verbose) log_printf("Open #%u: \"$%s\"", channel, path != NULL ? path : "");
    } else {
        Petscii name[17], type[4] = {'*', 0};
        const Petscii *outname;
        unsigned char overwrite;
        const char *path = NULL;
        int found = 0;
        int b;
        Directory_entry dirent;
        Directory *directory;
        Partition *outpart;

        buffer->mode = CM_CLOSED;
        overwrite = 0;
        for (b = 0; cmd[b]; b++) {
            if (cmd[b] != ':') continue;
            if (cmd[0] == '@') overwrite = 1;
            break;
        }

        outname = resolv_part(overwrite ? (cmd + 1) : cmd, &outpath, &outpart, 0);
        if (outname == NULL) {
            if (verbose) log_printf("Open #%u: Partition not found", channel);
            status = ER_DRIVE_NOT_READY; goto vege;
        }
        outname = resolv_path(outname, &outpath, arguments->nameconversion);
        path = outpath.data;
        if (outname == NULL) {
            status = ER_PATH_NOT_FOUND; goto vege2;
        }
        if (channel == 1) memcpy(type, "PRG", 4);
        convertfilename(outname, ',', name, type, &mode);
        if (channel == 0) mode = 'R';
        if (channel == 1) mode = 'W';

        if (!name[0]) {
            status = ER_MISSING_FILENAME; goto vege2;
        }

        directory = directory_open(&outpath, arguments->nameconversion, 0);
        if (directory == NULL) {
            status = directory_error(path);
            goto vege;
        }

        while (directory_read(directory, &dirent))
        {
            if ((dirent.attrib & (A_ANY | A_CLOSED)) != (A_CLOSED | A_NORMAL)) continue;

            if (!matchname(dirent.name, name)) continue;
            if (!matchname(dirent.filetype, type)) continue;

            path = path_append(&outpath, directory_filename(directory));
            found = 1;
            break;
        }
        directory_close(directory);

        if (!found) {
            Petscii lname[MAX_C64NAME_LEN];
            convertc64name(lname, name, type, arguments->nameconversion);
            path = path_appendc64(&outpath, lname);
            if (mode == 'W' && invalid_c64name(lname)) {
                status = ER_INVALID_FILENAME; goto vege2;
            }
        }

        if (path == NULL) {
            status = ER_READ_ERROR;
            log_print("Open: Out of memory");
            goto vege2;
        }

        if (mode == 'W') {
            length = 0;

            if (found && !overwrite) {
                status = ER_FILE_EXISTS;
            } else {
#if defined _WIN32 || defined __DJGPP__
                int fd = utf8_open(path, O_CREAT | O_RDWR | (overwrite ? O_TRUNC : O_EXCL) | O_BINARY, S_IRUSR | S_IWUSR);
#else
                int fd = utf8_open(path, O_CREAT | O_RDWR | (overwrite ? O_TRUNC : O_EXCL), S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif
                if (fd < 0) {
                    status = errnotochannel15(errno);
                } else {
                    buffers_close_name(path);
                    if (buffer_open(buffer, fd, path, "wb+")) {
                        status = errnotochannel15(errno);
                        close(fd);
                    } else {
                        buffer->mode = CM_FILE;
                        buffer->filesize = 0;
                        status = ER_OK;
                    }
                }
            }
        } else {
            if (!found) {
                status = ER_FILE_NOT_FOUND;
            } else {
                int fd = utf8_open(path, (mode == 'R') ? (O_RDONLY | O_BINARY) : (O_RDWR | O_BINARY), 0);
                if (fd < 0) status = errnotochannel15(errno); else {
#ifdef _WIN32
                    length = dirent.size;
#else
                    struct stat st;
                    int ret;
                    while ((ret = fstat(fd, &st)) == -1 && errno == EINTR);
                    length = (ret == -1) ? 0 : (st.st_size & ~(off_t)~0U) == 0 ? (unsigned int)st.st_size : ~0U;
#endif
                    buffers_close_name(path);
                    if (buffer_open(buffer, fd, path, (mode == 'R') ? "rb" : "rb+")) {
                        status = errnotochannel15(errno);
                        close(fd);
                    } else {
                        buffer->mode = CM_FILE;
                        buffer->filesize = length;
                        status = ER_OK;
                    }
                }
            }
        }
    vege2:
        if (verbose) log_printf("Open #%u: \"%s\" %c", channel, path != NULL ? path : "", mode);
    }
vege:
    path_free(&outpath);
    if (verbose) {
        log_time("Open: Took", &start);
    }
    {
        unsigned char buf[6];
        if (status == ER_UNKNOWN_ERROR) status = ER_READ_ERROR;
        if (driver->mode == M_ETHERNET) {
            buf[0] = (unsigned char)length;
            buf[1] = (unsigned char)(length >> 8);
            buf[2] = (unsigned char)(length >> 16);
            buf[3] = (unsigned char)(length >> 24);
            buf[4] = mode;
            buf[5] = (unsigned char)(status | 0x80);
        } else {
            buf[0] = (unsigned char)(status | 0x80);
            buf[1] = mode;
            buf[2] = (unsigned char)length;
            buf[3] = (unsigned char)(length >> 8);
            buf[4] = (unsigned char)(length >> 16);
            buf[5] = (unsigned char)(length >> 24);
            if (usecrc) {
                crc_clear();
                crc_add_block(buf, sizeof buf);
            }
        }
        driver->sendbytes(buf, sizeof buf);
        return send_trailer(driver, usecrc);
    }
}

int closefile(const Driver *driver, int unsigned flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    unsigned int length;
    struct timeval start;
    Buffer *buffer;
    unsigned char buf[7];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 5 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    length = (((((buf[4] * 0x100u) | buf[3]) * 0x100u) | buf[2]) * 0x100u) | buf[1];
    if (driver->done() != DRIVER_SUCCESS) return 1;
    if (driver->mode != M_ETHERNET) {
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x5a) return frame_error("Close");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Close");
        }
    }
    buffer = buffers_get(channel);

    if (verbose) {
        gettimeofday(&start, NULL);
    }

    if (verbose) log_printf("Close #%u: %d byte(s)", channel, length);
    if (channel == 15) buffer->mode = CM_ERR;
    else {
        if (buffer->file != NULL) {
            if (buffer->mode == CM_FILE && length >= buffer->filesize) {
                fflush(buffer->file);
                if (ftruncate(buffer->fd, length)) {
                    log_printf("Close: Couldn't truncate: %s(%d)", strerror(errno), errno);
                }
            }
            buffer_close(buffer);
        }
        buffer->mode = CM_CLOSED;
        buffer_free(buffer);
    }
    if (verbose) {
        log_time("Close: Took", &start);
    }
    driver->sendb(0x80 | ER_OK);
    if (driver->mode == M_ETHERNET) driver->sendb(0x80 | ER_OK);
    else if (usecrc) {
        crc_first(0x80 | ER_OK);
    }
    return send_trailer(driver, usecrc);
}

int readfile(const Driver *driver, int unsigned flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    unsigned char sectors;
    struct timeval start;
    Errorcode status = ER_OK;
    unsigned int l = 0;
    unsigned int address;
    Buffer *buffer;
    unsigned char buf[7];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 5 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    address = (((buf[3] * 0x100u) | buf[2]) * 0x100u) | buf[1];
    sectors = buf[4];
    if (driver->done() != DRIVER_SUCCESS) return 1;
    if (driver->mode != M_ETHERNET) {
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x5a) return frame_error("Read");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Read");
        }
    }
    buffer = buffers_get(channel);

    if (verbose) log_printf("Read #%u: %06x %d sector(s)", channel, address, sectors);
    if (sectors > 128 || sectors == 0 || (driver->mode == M_ETHERNET && sectors > 2)) {
        status = ER_READ_ERROR;
        log_print("Read: Invalid sector count");
        goto error;
    }
    if (buffer->mode != CM_DIR && buffer->mode != CM_ERR && (buffer->mode != CM_FILE || buffer->file == NULL)) {
        status = ER_NO_CHANNEL;
        log_print((buffer->mode == CM_CLOSED) ? "Read: Channel not open" : "Read: Not readable");
    error:
        if (status == ER_UNKNOWN_ERROR) status = ER_READ_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        if (driver->mode == M_ETHERNET) driver->sendb((unsigned char)(0x80 | status));
        else if (usecrc) {
            crc_first((unsigned char)(0x80 | status));
        }
        return send_trailer(driver, usecrc);
    }

    if (buffer->mode == CM_FILE) {
        if (buffer_reserve(buffer, (driver->mode == M_ETHERNET) ? sectors * 512u : 512)) {
            status = ER_READ_ERROR;
            log_print("Read: Out of memory");
            goto error;
        }
        if (buffer->filepos != address) {
            if (fseek(buffer->file, address << 8, SEEK_SET) != 0) {
                buffer->filepos = ~0u;
                status = ER_READ_ERROR;
                log_printf("Read: Couldn't seek: %s(%d)", strerror(errno), errno);
                goto error;
            }
            buffer->filepos = address;
        }
        clearerr(buffer->file);
    }

    if (driver->mode == M_ETHERNET) {
        const unsigned char *data;
        if (buffer->mode == CM_FILE) {
            size_t itt = buffer_read(buffer, sectors * 512u);
            if (itt != sectors * 512u) {
                if ((itt & 511) != 0) {
                    buffer->filepos = ~0u;
                } else {
                    buffer->filepos += itt >> 8;
                }
                memset(buffer->data + itt, 0, sectors * 512u - itt);
                if (ferror(buffer->file)) {
                    status = errnotochannel15(errno);
                    log_printf("Read: Couldn't read: %s(%d)", strerror(errno), errno);
                }
            } else {
                buffer->filepos += sectors * 2u;
            }
            data = buffer->data;
        } else {
            data = buffer->data + buffer->pointer;
        }
        driver->sendbytes(data, sectors * 512u);
        if (status == ER_UNKNOWN_ERROR) status = ER_READ_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        driver->sendb((unsigned char)(0x80 | status));
        if (driver->flush() != DRIVER_SUCCESS) return 1;
        buffer->pointer += sectors * 512u;
    } else {
        if (status == ER_UNKNOWN_ERROR) status = ER_READ_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        if (usecrc) {
            crc_first((unsigned char)(0x80 | status));
            driver->sendb((unsigned char)crc_get());
        }
        driver->sendb(0x5a);
        if (driver->mode == M_RS232 || driver->mode == M_RS232S) {
            if (driver->flush() != DRIVER_SUCCESS) return 1;
        }
        if (verbose) {
            gettimeofday(&start, NULL);
        }
        while (sectors) {
            int err = 0;
            const unsigned char *data;
            if (buffer->mode == CM_FILE) {
                size_t itt = buffer_read(buffer, 512);
                if (itt != 512) {
                    if (itt != 0) buffer->filepos = ~0u;
                    memset(buffer->data + itt, 0, 512 - itt);
                    err = ferror(buffer->file);
                    if (err) {
                        status = errnotochannel15(errno);
                        log_printf("Read: Couldn't read: %s(%d)", strerror(errno), errno);
                    }
                } else {
                    buffer->filepos += 2;
                }
                data = buffer->data;
            } else {
                data = buffer->data + buffer->pointer;
            }
            if (driver->mode == M_RS232 || driver->mode == M_RS232S) {
                int fr = driver->getb(1);
                if (driver->done() != DRIVER_SUCCESS) return 1;
                if (fr != 0x5a) return frame_error("Read");
            }
            driver->sendbytes(data, 512);
            if (status == ER_UNKNOWN_ERROR) status = ER_READ_ERROR;
            driver->sendb((unsigned char)(0x80 | status));
            if (usecrc) {
                crc_clear();
                crc_add_block(data, 512);
                crc_add_byte((unsigned char)(0x80 | status));
                driver->sendb((unsigned char)crc_get());
            }
            driver->sendb(0x5a);
            if (driver->mode == M_RS232 || driver->mode == M_RS232S) {
                if (driver->flush() != DRIVER_SUCCESS) return 1;
            }
            if (err) break;
            buffer->pointer += 512;
            sectors--; l += 512;
        }
        if (driver->mode != M_RS232 && driver->mode != M_RS232S) {
            if (driver->flush() != DRIVER_SUCCESS) return 1;
        }
        if (verbose) {
            log_speed("Read: Sent", l, &start);
        }
    }
    return 0;
}

int writefile(const Driver *driver, int unsigned flags) {
    const int usecrc = flags & FLAG_USE_CRC;
    const int usepadding = flags & FLAG_USE_PADDING;
    const int verbose = flags & FLAG_VERBOSE;
    unsigned int channel;
    unsigned char sectors;
    struct timeval start;
    Errorcode status = ER_OK;
    unsigned int l = 0;
    unsigned int address;
    Buffer *buffer;
    unsigned char buf[7];

    driver->getbytes(buf, driver->mode == M_ETHERNET ? 5 : usecrc ? sizeof buf : sizeof buf - 1);
    channel = buf[0] & 0x0f;
    address = (((buf[3] * 0x100u) | buf[2]) * 0x100u) | buf[1];
    sectors = buf[4];
    if (driver->done() != DRIVER_SUCCESS) return 1;
    if (driver->mode != M_ETHERNET) {
        if (buf[usecrc ? sizeof buf - 1 : sizeof buf - 2] != 0x5a) return frame_error("Write");
        if (usecrc) {
            crc_add_block(buf, sizeof buf - 1);
            if (crc_get()) return crc_error("Write");
        }
    }
    buffer = buffers_get(channel);

    if (verbose) log_printf("Write #%u: %06x %d sector(s)", channel, address, sectors);
    if (sectors > 128 || sectors == 0 || (driver->mode == M_ETHERNET && sectors > 2)) {
        status = ER_WRITE_ERROR;
        log_print("Write: Invalid sector count");
        goto error;
    }
    if (buffer->mode != CM_FILE || buffer->file == NULL) {
        status = ER_NO_CHANNEL;
        log_print("Write: Not writeable");
    error:
        if (status == ER_UNKNOWN_ERROR) status = ER_WRITE_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        if (driver->mode == M_ETHERNET) driver->sendb((unsigned char)(0x80 | status));
        else if (usecrc) {
            crc_first((unsigned char)(0x80 | status));
        }
        return send_trailer(driver, usecrc);
    }

    if (buffer_reserve(buffer, (driver->mode == M_ETHERNET) ? 512u * sectors : 514)) {
        status = ER_WRITE_ERROR;
        log_print("Write: Out of memory");
        goto error;
    }
    if (buffer->filepos != address) {
        if (fseek(buffer->file, address << 8, SEEK_SET) != 0) {
            buffer->filepos = ~0u;
            status = ER_WRITE_ERROR;
            log_printf("Write: Couldn't seek: %s(%d)", strerror(errno), errno);
            goto error;
        }
        buffer->filepos = address;
    }
    if (driver->mode == M_ETHERNET) {
        driver->getbytes(buffer->data, 512u * sectors);
        if (driver->done() != DRIVER_SUCCESS) return 1;
        if (buffer_write(buffer, 512u * sectors) == 512 * sectors && fflush(buffer->file) == 0) {
            buffer->filepos += sectors * 2u;
        } else {
            buffer->filepos = ~0u;
            status = errnotochannel15(errno);
            log_printf("Write: Couldn't write: %s(%d)", strerror(errno), errno);
        }
        if (status == ER_UNKNOWN_ERROR) status = ER_WRITE_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        driver->sendb((unsigned char)(0x80 | status));
        if (driver->flush() != DRIVER_SUCCESS) return 1;
    } else {
        int err = 0;
        if (status == ER_UNKNOWN_ERROR) status = ER_WRITE_ERROR;
        driver->sendb((unsigned char)(0x80 | status));
        if (usecrc) {
            crc_first((unsigned char)(0x80 | status));
        }
        if (send_trailer(driver, usecrc)) return 1;
        if (sectors > 1) driver->turn(-1);
        if (verbose) {
            gettimeofday(&start, NULL);
        }
        while (sectors) {
            if (sectors == 1) driver->turn(0x5a);
            driver->getbytes(buffer->data, (usepadding || usecrc) ? 514 : 513);
            if (driver->done() != DRIVER_SUCCESS) return 1;
            if (buffer->data[(usepadding || usecrc) ? 513 : 512] != 0x5a) return frame_error("Write");
            if (usecrc) {
                crc_clear();
                crc_add_block(buffer->data, 513);
                if (crc_get()) return crc_error("Write");
            }
            if (err == 0 && buffer_write(buffer, 512) != 512) err = errno;
            buffer->filepos += 2;
            sectors--; l += 512;
            if (sectors == 0 || driver->mode == M_RS232 || driver->mode == M_RS232S) {
                if (err == 0 && fflush(buffer->file) != 0) err = errno;
                if (err != 0) {
                    buffer->filepos = ~0u;
                    status = errnotochannel15(err);
                    log_printf("Write: Couldn't write: %s(%d)", strerror(err), err);
                }
                if (status == ER_UNKNOWN_ERROR) status = ER_WRITE_ERROR;
                driver->sendb((unsigned char)(0x80 | status));
                if (usecrc) {
                    crc_first((unsigned char)(0x80 | status));
                }
                if (send_trailer(driver, usecrc)) return 1;
                if (err) break;
            }
        }
        if (verbose) {
            log_speed("Write: Received", l, &start);
        }
    }
    return 0;
}

